#include "copyright.h"
/*
 * These are subroutines to read unformatted files, given a
 * description of their vaguaries in a `filecap' file. The files are
 * permitted to have records which are padded either at the front or
 * the back (or both) with extraneous bytes, to have a padding in front
 * of the entire file, and to have a header. Various fields are extracted
 * from this header, if requested in the filecap file.
 *
 * Filecap capabilities used are :
 *
 * DA           DAta type       char fits float(D) int long short (string)
 * FS           File Start      Unwanted bytes at start of file
 * HS           Header Size     Size of header (alias: HH)
 * NS           No Swap         Don't swap bytes, even on (e.g.) a vax
 * RE           Record End      Unwanted bytes at end of record
 * RL           Record Length   Number of useful bytes per record
 * RS           Record Start    Unwanted bytes at start of record
 * SW		SWap		Swap bytes (ab --> ba, abcd --> dcba)
 * nx,ny        Number X, Y     byte offsets of X and Y sizes of data
 *                              within the first record, sizes are ints
 *
 * Most parameters are optional, and will default to 0. If RE or RS is
 * specified, you must give RL as well. If you specify it as -ve, then
 * we'll look for it from the O/S.
 * Nx and ny must be present. Note that HH excludes FS, so HH will usually
 * be 2*sizeof(int) irrespective of the value of FS, and nx will usually be 0.
 */
#include "options.h"
#include <stdio.h>
#include <ctype.h>
#include "sm.h"
#include "image.h"
#ifdef vms
#  include <stat.h>
#else
#  include <sys/types.h>
#  include <sys/stat.h>
#endif
#include "tty.h"
#include "tree.h"
#include "sm_declare.h"

#define FILECAP_SIZE 200		/* buffer for list of filecaps */
#define FITS_SIZE 2880
#define SSIZE 40                        /* size for small buffers */
extern int sm_interrupt,                   /* respond to ^C */
	   sm_verbose;                     /* control volubility */

#define CONVERT                 /* convert cbuff to float in zz */      \
   { register int jj;                                                   \
     switch(dtype) {                                                    \
       case CHAR:                                                       \
	 for(jj = 0;jj < rsize;jj++)                                    \
	    *((float *)&zz[zi] + jj) = *(cbuff+jj); break;              \
       case INT:                                                        \
	 for(jj = 0;jj < rsize/sizeof(int);jj++)                        \
	    *((float *)&zz[zi] + jj) = *((int *)cbuff + jj); break;     \
       case LONG:                                                       \
	 for(jj = 0;jj < rsize/sizeof(long);jj++)                       \
	    *((float *)&zz[zi] + jj) = *((long *)cbuff + jj); break;    \
       case SHORT:                                                      \
	 for(jj = 0;jj < rsize/sizeof(short);jj++)                      \
	    *((float *)&zz[zi] + jj) = *((short *)cbuff + jj); break;   \
       default:                                                         \
	 msg("Impossible dtype in get_image!\n"); abort();                \
     }}
#define BITS_IN_CHAR 8                  /* convert FITS BITPIX to sizeof */
#define CHAR 0                          /* symbolic names for dtype */
#define FLOAT 1
#define INT 2
#define LONG 3
#define SHORT 4
#define SIZE 10240              /* size of buffer to use in simple cases */

static char *get_capstr(),
	    *get_keyword(),
	    *header = NULL;		/* storage for header entry */
extern int sm_interrupt,                   /* respond to ^C */
	   sm_verbose;                     /* control volubility */
static int find_entry(),
	   get_capint(),
	   get_recl(),
	   header_size = 0,		/* size of header */
	   is_capval(),
	   read_card(),
	   read_fits_header();
static void restore_region(),
	    save_region();
static void swap_2(),
	    swap_4();

IMAGE *
get_image(filename)
char *filename;
{
   char arr[SIZE],                      /* array for header */
   	buff[30],			/* temp. buffer */
	*cbuff = NULL,			/* buffer for converting to floats
					   (initialised to NULL to placate
					   flow-analysing compilers) */
	*data_type,                     /* type of data */
	filecap[FILECAP_SIZE],		/* name of filecap files */
	*file_type,                     /* name of entry in filecap */
	*ptr,
	*zz;                            /* storage for z */
   float bscale =  0.0,                 /* if non-zero, */
	 bbzero = 0.0,                  /*  data = bbzero + bscale*data */
	 xmin,xmax,                     /* range of x and */
	 ymin,ymax,                     /*        y coordinates */
	 **z;                           /* pointers to image */
   IMAGE *image;                        /* the image to create */
   int dsize,                           /* size of each datum */
       dtype,                           /* type of data in file */
       fd = -1,                         /* file descriptor for data */
       fits = 0,                        /* is it a FITS image? */
       i,				/* index, esp. to data as read */
       file_start,                      /* dead space at start of file */
       head_size,                       /* size of header */
       head_start,			/* offset of start of header */
       have_xlims = 0,have_ylims = 0,   /* were x/y limits on header? */
       swap_bytes = 0,			/* do we want to byte swap? */
       n_read,				/* number of bytes still to read */
       nx,ny,                           /* dimension of image */
       rec_start,                       /* dead space at start of record */
       rec_len,                         /* length of data per record */
       rec_end,                         /* dead space at end of record */
       rsize,                           /* number of bytes to read at a time */
       size,                            /* space needed for data */
       zi;                              /* index into z; != i if not floats */

   kill_keywords();			/* kill old keywords from header */
   if((fd = open(filename,0)) < 0) {
      msg_1s("Can't open %s\n",filename);
      return(NULL);
   }
/*
 * Find the `filecap' file, then the characteristics of file,
 * then read the header and then the data itself
 * Note that the filecap entries will usually be in the graphcap file,
 * and this will be assumed if we can't find filecap.
 */
   if((ptr = get_val("filecap")) == NULL) {
      if((ptr = get_val("graphcap")) == NULL) {
	 msg("Can't find filecap/graphcap in environment file\n");
	 return(NULL);
      }
   }
   (void)strncpy(filecap,ptr,FILECAP_SIZE);
   file_type = print_var("file_type");
   if(file_type[0] == '\0') {
      msg("No variable file_type is defined, assuming \"C\"\n");
      find_entry("C",filecap);
   } else {
      find_entry(file_type,filecap);
   }

   head_start = file_start = get_capint("FS");
   if((head_size = get_capint("HS")) == 0) {
      head_size = get_capint("HH");     /* old name for it */
   }
   rec_start = get_capint("RS");
   rec_len = get_capint("RL");
   rec_end = get_capint("RE");
   data_type = get_capstr("DA");
   if(*data_type == '\0') data_type = "float";  /* as a default */
/*
 * process values, looking for errors and special values.
 *
 * If HS is -ve, assume that the file has a record structure
 * that even applies to the header. If RL is specified, use it for HS.
 * Otherwise get it from the O/S.
 *
 * If RL is -ve, get it from the O/S
 */
   if(sm_verbose > 0) {
      msg_1s("Data type is %s\n",data_type);
   }
   if(!strcmp(data_type,"char")) {
      dtype = CHAR; dsize = sizeof(char);
   } else if(!strcmp(data_type,"fits")) {
      fits = 1;
      if(rec_len != 0 && rec_len != FITS_SIZE) {
	 msg_1d("FITS files must have a record length of %d\n",FITS_SIZE);
      }
      head_size = rec_len = FITS_SIZE;
      head_size += rec_start;           /* fake up the header */
      head_size += rec_end;
      head_start += rec_start;
      dtype = INT;			/* we'll check in read_fits_header() */
   } else if(!strcmp(data_type,"float")) {
      dtype = FLOAT; dsize = sizeof(float);
   } else if(!strcmp(data_type,"int")) {
      dtype = INT; dsize = sizeof(int);
   } else if(!strcmp(data_type,"long")) {
      dtype = LONG; dsize = sizeof(long);
   } else if(!strcmp(data_type,"short")) {
      dtype = SHORT; dsize = sizeof(short);
   } else {
      msg_1s("Data type %s not supported\n",data_type);
      (void)close(fd);
      return(NULL);
   }

#ifdef NEED_SWAP			/* we need to byte swap FITS */
   if(fits) swap_bytes = !is_capval("NS"); /* unless explicitly forbidden */
#endif
   if(is_capval("SW")) swap_bytes = 1;
   if(swap_bytes) {
      if(dtype == SHORT || dtype == INT || dtype == LONG) {
	 ;				/* OK */
      } else {
	 msg_1s("You cannot request byte swapping for data type %s\n",
		data_type);
	 swap_bytes = 0;
      }
   }

   if(head_size < 0) {
      if(rec_len > 0) {
	 head_size = rec_len;
      } else {
	 if((head_size = get_recl(fd)) < 0) {
	    msg_1s("Can't get record length for %s\n",filename);
	    close(fd);
	    return(NULL);
	 }
      }
      head_size += rec_start;           /* fake up the header */
      head_size += rec_end;
      head_start += rec_start;
   }
   if(rec_start > SSIZE) {
      msg_1d("rec_start is too large, > %d\n",SSIZE);
      (void)close(fd);
      return(NULL);
   }
   if((rec_start || rec_end) && !rec_len) {
      msg("You must specify RL in filecap if you use RS or RE; Assuming -1\n");
      rec_len = -1;
   }
/*
 * Read header
 */
   if(fits) {                   /* we already have rec_len */
      if((size = file_start + head_size) > SIZE) {
	 msg("Array in get_image is too small to read header\n");
	 (void)close(fd);
	 return(NULL);
      }
      ptr = arr; 
      for(i = rec_len;i < size;i += rec_len) {
	 if(read(fd,ptr,rec_len) != rec_len) {
	    msg("Error reading header\n");
	    (void)close(fd);
	    return(NULL);
	 }
	 ptr += rec_len;     
      } 
      n_read = size - (ptr - arr);	/* this many by of size still unread */
      if(read(fd, ptr, n_read) != n_read) {
	 msg("Error reading end of header\n");
	 (void)close(fd);
	 return(NULL);
      }
/*
 * Process the header
 */
      nx = 0;                           /* this is the first header record */
      for(;(i = read_fits_header(&arr[head_start],&nx,&ny,
		&dtype,&dsize,&bscale,&bbzero,&xmin,&xmax,&ymin,&ymax)) != 0;){
	 if(i < 0) {                    /* error on header */
	    (void)close(fd);
	    return(NULL);
	 } else {                       /* not seen END card yet */
	    size -= file_start;
	    file_start = 0;             /* we've read file_start already */
	    n_read = rec_len - n_read;	/* How many bytes we must still read 
                                           from current record */
	    ptr = arr;
	    for(i = n_read; i < size ;i += rec_len) {
	       if(read(fd, ptr, n_read) != n_read) {
		  msg("Error reading header\n");
		  (void)close(fd);
		  return(NULL);
	       }
	       ptr += n_read;
	       n_read = rec_len;
	    } 
	    n_read = size - (ptr -arr);
	    if((i = read(fd, ptr, n_read)) != n_read) {
	       msg("Error reading header\n");
	       (void)close(fd);
	       return(NULL);
	    }
	    ptr += n_read;
	 }
	 if(xmin != 0 || xmax != nx - 1) have_xlims = 1;
	 if(ymin != 0 || ymax != ny - 1) have_ylims = 1;
      }
   } else {                             /* Not FITS */
      if((size = file_start + head_size) > SIZE) {
	 msg("Array in get_image is too small to read header\n");
	 (void)close(fd);
	 return(NULL);
      }
      ptr = arr;
      if(rec_len <= 0) {
	 if(read(fd, ptr, size) != size) {
	    msg("Error reading header\n");
	    (void)close(fd);
	    return(NULL);
	 }
	 n_read = 0;			/* num of bytes of 1st record unread */
      } else {
	 for(i = 0;i < size - rec_len;i += rec_len) {
	    if(read(fd, ptr, rec_len) != rec_len) {
	       msg("Error reading header\n");
	       (void)close(fd);
	       return(NULL);
	    }
	    ptr += rec_len;
	 }
	 n_read = size - (ptr - arr);	/* num of bytes of 1st record unread */
      }
      if((i = read(fd, ptr, n_read)) != n_read) {
	 msg("Error reading header\n");
	 (void)close(fd);
	 return(NULL);
      }
      
      if(head_size < 200) {		/* save the header */
	 if(header != NULL) free(header);
	 if((header = (char *)malloc(head_size)) != NULL) {
	    memcpy(header,arr + file_start,head_size);
	    header_size = head_size;
	 }
      }
      if(!is_capval("nx") && !is_capval("ny")) {
	 prompt_user("Enter nx and ny ");
	 if(mgets(arr,SIZE) == NULL || str_scanf(arr,"%d %d",&nx,&ny) != 2) {
	    msg("Error reading size of image interactively\n");
	    (void)close(fd);
	    return(NULL);
	 }
      } else {
	 if(!is_capval("nx") || !is_capval("ny")) {
	    msg("Size of image is not present in header\n");
	    (void)close(fd);
	    return(NULL);
	 }
/*
 * I hope that the values in filecap won't give an alignment problem
 */
	 if(*(ptr = get_capstr("nx")) != '\0' && *ptr != '#') {
	    nx = atoi(ptr);		/* the value of nx */
	    if(sm_verbose > 1) msg_1d("nx is %d\n",nx);
	 } else {
	    nx = get_capint("nx");
	    if(nx < 0) {
	       if(sm_verbose > 1) msg_1d("nx is %d, ",-nx);
	       nx = -nx;
	    } else {
	       if(sm_verbose > 1) msg_1d("nx is at offset %d, ",head_start + nx);
	       nx = *(int *)&arr[head_start + nx];
	    }
	 }
	 
	 if(*(ptr = get_capstr("ny")) != '\0' && *ptr != '#') {
	    ny = atoi(ptr);		/* the value of ny */
	    if(sm_verbose > 1) msg_1d("ny is %d\n",ny);
	 } else {
	    ny =  get_capint("ny");
	    if(ny < 0) {
	       if(sm_verbose > 1) msg_1d("ny is %d\n",-ny);
	       ny = -ny;
	    } else {
	       if(sm_verbose > 1) msg_1d("ny is at offset %d\n",head_start + ny);
	       ny = *(int *)&arr[head_start + ny];
	    }
	 }
      }
/*
 * check that nx and ny are reasonable. Say, >= 0 && size <= 1e8
 */
      if(nx <= 0 || ny <= 0) {
	 msg_1d("You have nx = %d, ",nx);
	 msg_1d("ny = %d\n",ny);
	 (void)close(fd);
	 return(NULL);
      }
      if((unsigned)nx*ny > 1e8) {    /* probably junk */
	 msg_1d("You have nx = %d, ",nx);
	 msg_1d("ny = %d, is this correct?\n",ny);
	 msg("enter 1 to continue  ");
	 (void)scanf("%d%*c",&i);
	 if(!i) {
	    (void)close(fd);
	    return(NULL);
	 }
      }
   }
   (void)sprintf(buff,"%d",nx); set_image_variable("NX",buff);
   (void)sprintf(buff,"%d",ny); set_image_variable("NY",buff);
/*
 * look for x/y limits of image,
 * and hope that the values in filecap won't give an alignment problem
 */
   if(is_capval("x0")) {
      if(!is_capval("x1")) {
	 msg("You must specify neither or both of x0 and x1 in a file\n");
      } else {
	 xmin = *(float *)&arr[file_start + rec_start + get_capint("x0")];
	 xmax = *(float *)&arr[file_start + rec_start + get_capint("x1")];
	 (void)sprintf(buff,"%g",xmin); set_image_variable("X0",buff);
	 (void)sprintf(buff,"%g",xmax); set_image_variable("X1",buff);
	 have_xlims = 1;
      }
   }
   if(is_capval("y0")) {
      if(!is_capval("y1")) {
	 msg("You must specify neither or both of y0 and y1 in a file\n");
      } else {
	 ymin = *(float *)&arr[file_start + rec_start + get_capint("y0")];
	 ymax = *(float *)&arr[file_start + rec_start + get_capint("y1")];
	 (void)sprintf(buff,"%g",ymin); set_image_variable("Y0",buff);
	 (void)sprintf(buff,"%g",ymax); set_image_variable("Y1",buff);
	 have_ylims = 1;
      }
   }

   if(sm_verbose > 0) {
      msg_1s("Image %s's ",filename);
      msg_1d("dimensions are %d*",nx);
      msg_1d("%d\n",ny);
      if(have_xlims) {
	 msg_1f("x: %g - ",xmin);
	 msg_1f("%g\n",xmax);
      }
      if(have_ylims) {
	 msg_1f("y: %g - ",ymin);
	 msg_1f("%g\n",ymax);
      }
   }
/*
 * Now get the record length if we were asked for it. We couldn't get it
 * before because the first record (with header information) can have a
 * different value of rec_len
 */
   if(rec_len < 0) {            /* must get it from the O/S */
      if((rec_len = get_recl(fd)) < 0) {
	 msg_1s("Can't get record length for %s\n",filename);
	 close(fd);
	 return(NULL);
      }
   }
/*
 * Now we can read the real data. We must allow for the record structure
 * when reading it.
 */
   size = nx*ny*dsize;
   rsize = rec_start + rec_len + rec_end;
   if(rsize == 0) {             /* no record structure is specified */
      rsize = (size > SIZE) ? SIZE : size;
   }
   if(size <= 0) {
      msg("Non-positive storage requested for image\n");
      (void)close(fd);
      return(NULL);
   }
/*
 * Now get storage for data, and pointers to data
 * Note that we start reading at &zz[0], but that the first rec_start of
 * the record is junk. Allow for this by moving the z[] pointers up.
 * If we have to convert we convert to a different format we have to
 * allow for the space taken up by converting rec_start of data to float
 */
   if((zz = malloc(rec_start + nx*ny*sizeof(float))) == NULL) {
      msg("Can't allocate image in get_image\n");
      (void)close(fd);
      return(NULL);
   }
   if(dtype != FLOAT) {                 /* also need conversion buffer */
      if((cbuff = malloc(rsize)) == NULL) {
	 msg("Can't allocate cbuff in get_image\n");
	 free(zz); zz = NULL;
	 (void)close(fd);
	 return(NULL);
      }
   }
/*
 * allocate and set up the pointers to the image
 */
   if((z = (float **)malloc(ny*sizeof(float *))) == NULL) {
      msg("Can't allocate image pointers in get_image\n");
      free(zz); zz = NULL;
      if(dtype != FLOAT) free(cbuff);
      (void)close(fd);
      return(NULL);
   }
   for(i = 0;i < ny;i++) {
      z[i] = (float *)(zz + rec_start) + i*nx;
   }
/*
 * We'll now simply read the data if rec_start == 0, otherwise
 * save a few bytes in a temp space to avoid being overwritten by it.
 */
   if(rec_start == 0) {
      for(zi = i = 0;!sm_interrupt && i < size;) {
	 if(rsize > size - i) {         /* may be junk at end of file */
	    rsize = size - i;
	 }
	 if(dtype == FLOAT) {
	    if((rsize = read(fd,&zz[zi],rsize)) <= 0) {
	       break;
	    }
	 } else {
	    if((rsize = read(fd,cbuff,rsize)) <= 0) {
	       break;
	    }
	    if(swap_bytes) {
	       if(dsize == 2) {
		  swap_2(cbuff,rsize);
	       } else if(dsize == 4) {
		  swap_4(cbuff,rsize);
	       }
	    }
	    CONVERT;
	 }
	 i += rsize - rec_end;
	 zi += (rsize - rec_end)*sizeof(float)/(float)dsize;
      }
      if(i < size) {
	 i += rsize - rec_end;          /* how many bytes we read */
      }
   } else {
/*
 * Note that we start reading at &zz[0], but that the first rec_start of
 * the record is junk. We allowed for this by moving the z[] pointers up.
 */
      for(zi = i = 0;!sm_interrupt && i < size;) {
	 save_region(&zz[zi],rec_start);        /* save rec_start region */
	 if(rsize > size + rec_start - i) {     /* may be junk at end of file*/
	    rsize = size + rec_start - i;
	 }
	 if(dtype == FLOAT) {
	    if((rsize = read(fd,&zz[zi],rsize)) <= 0) {
	       break;
	    }
	 } else {
	    if((rsize = read(fd,cbuff,rsize)) <= 0) {
	       break;
	    }
	    if(swap_bytes) {
	       if(dsize == 2) {
		  swap_2(&cbuff[rec_start],rsize - rec_start);
	       } else if(dsize == 4) {
		  swap_4(&cbuff[rec_start],rsize - rec_start);
	       }
	    }
	    CONVERT;
	 }
	 restore_region(&zz[zi],rec_start);     /* and restore it */
	 i += rec_len;
	 zi += rec_len*sizeof(float)/(float)dsize;
      }
      if(i < size) {
	 i += rsize - rec_start;        /* how many bytes we read */
      }
   }
   if(i < size) {                       /* we didn't find enough data */
      msg_1s("Data in file %s ",filename);
      msg_1d("contained %d ",i);
      msg_1d("not %d bytes\n",size);
   } else {                             /* check that file is all read */
      if((i = read(fd,arr,SIZE)) > rec_end) {
	 if(fits && i - rec_end < FITS_SIZE) {
	    ;				/* just the end of the record */
	 } else {
	    msg_1d("At least %d bytes remain unread at end of data file\n",
							i - rec_end);
	 }
      }
   }

   (void)close(fd);
   if(dtype != FLOAT) {                 /* also free conversion buffer */
      free(cbuff);
   }
   if((image = (IMAGE *)malloc(sizeof(IMAGE))) == NULL) {
      fprintf(stderr,"Can't allocate storage for image structure\n");
      return(NULL);
   }
   strncpy(image->name,filename,sizeof(image->name) - 1);
   image->nx = nx;
   image->ny = ny;
   if(have_xlims) {                     /* from file header */
      image->xmin = xmin;
      image->xmax = xmax;
   } else {
      image->xmin = 0;
      image->xmax = nx - 1;
   }
   if(have_ylims) {
      image->ymin = ymin;
      image->ymax = ymax;
   } else {
      image->ymin = 0;
      image->ymax = ny - 1;
   }
   image->ptr = z;
   image->space = zz;
/*
 * scale image if neccesary
 */
   if(bscale != 0.0 || bbzero != 0.0) { /* it is */
      register float *zend,*zptr;

      for(zptr = z[0],zend = z[0] + nx*ny;zptr < zend;zptr++) {
	 *zptr = *zptr*bscale + bbzero;
      }
   }
   return(image);
}

char *
read_image_variable(name)
char *name;				/* name of variable to find */
{
   char buff[40];
   int offset;
   char *val;

   if(*(val = get_keyword(name)) != '\0') { /* found it */
      return(val);
   }
   if(header == NULL) {
      return("");
   }
   if(is_capval(name)) {
      if((offset = get_capint(name)) > header_size - sizeof(float)) {
	 msg_1s("Offset for variable %s is too large\n",name);
	 return("");
      }
      sprintf(buff,"%g",*(float *)(&header[offset]));
      set_image_variable(name,buff);
      return(get_keyword(name));
   }
   return("");
}

/*
 * Make an image, and define its variables
 */
int
create_image(nx,ny,xmin,xmax,ymin,ymax)
int nx,ny;			/* dimension of data array */
double xmin,xmax,ymin,ymax;	/* size of data array in user coords */
{
   char buff[30];			/* temp. buffer */
   int i;
   float *s_data;			/* storage for image */
   float **data;			/* pointers to image */
   
   if((s_data = (float *)malloc(nx*ny*sizeof(float))) == NULL) {
      fprintf(stderr,"Can't allocate storage for image\n");
      return(-1);
   }
   if((data = (float **)malloc(ny*sizeof(float *))) == NULL) {
      fprintf(stderr,"Can't allocate storage for image\n");
      free((char *)s_data);
      return(-1);
   }

   data[0] = s_data;
   for(i = 0;i < nx;i++) {
      data[0][i] = 0.0;
   }
   for(i = 1;i < ny;i++) {
      data[i] = s_data + nx*i;
      memcpy((char *)data[i],(char *)data[0],nx*sizeof(float));
   }
   
   if(sm_defimage(data,xmin,xmax,ymin,ymax,nx,ny) < 0) {
      free((char *)data);
      free((char *)s_data);
      return(-1);
   }
   set_image_name("interactive SM IMAGE");
   
   (void)sprintf(buff,"%d",nx); set_image_variable("NX",buff);
   (void)sprintf(buff,"%d",ny); set_image_variable("NY",buff);
   (void)sprintf(buff,"%g",xmin); set_image_variable("X0",buff);
   (void)sprintf(buff,"%g",xmax); set_image_variable("X1",buff);
   (void)sprintf(buff,"%g",ymin); set_image_variable("Y0",buff);
   (void)sprintf(buff,"%g",ymax); set_image_variable("Y1",buff);
   
   return(0);
}

/******************************************************/
static char temp[SSIZE];                /* save region */

static void
save_region(ptr,n)
char *ptr;
int n;
{
   char *end,
	*tptr;

   tptr = temp;
   end = tptr + n;

   while(tptr < end) *tptr++ = *ptr++;
}

static void
restore_region(ptr,n)
char *ptr;
int n;
{
   char *end,
	*tptr;

   tptr = temp;
   end = tptr + n;

   while(tptr < end) *ptr++ = *tptr++;
}

/*******************************************************/
/*
 * Get the record length of the current record
 */
static int
get_recl(fd)
int fd;                                 /* current file */
{
   int recl;
#ifndef unix
   struct stat stat_buff;       /* used by fstat to interrogate system */
#endif

#if !defined(VMS)
   (void)read(fd,(char *)&recl,sizeof(int)); /* first int is record length */
   (void)lseek(fd,-sizeof(int),1);
#else
   if(fstat(fd,&stat_buff) != 0) {
      (void)close(fd);
      return(-1);
   }
   recl = stat_buff.st_fab_mrs;         /* record length for VMS */
#endif
   return(recl);
}

/***************************************************/
/*
 * Now routines to administer `filecap' - actually just interfaces
 * to the general *cap routines used primarily by graphcap.
 */

static TTY *file_desc = NULL;		/* descriptor for filecap */

static int
find_entry(name,file)
char name[],
     file[];
{
   if(file_desc != NULL) free((char *)file_desc);
   if((file_desc = ttyopen(file,1,&name,(int (*)())NULL)) == NULL) {
      msg("ttyopen returns NULL in find_entry()\n");
      return(-1);
   }
   tty_index_caps(file_desc,file_desc->t_capcode,file_desc->t_capindex);
   return(0);
}

static int
get_capint(cap)
char *cap;
{
   if(file_desc == NULL) return(0);     /* file not correctly opened */
   return(ttygeti(file_desc,cap));
}

/*
 * Does capability exist?
 */
static int
is_capval(cap)
char *cap;
{
   if(file_desc == NULL) return(0);     /* file not correctly opened */
   return(ttygetb(file_desc,cap));
}


static char *
get_capstr(cap)
char *cap;
{
   static char capstr[21];

   if(file_desc == NULL) return("");  /* file not correctly opened */
   if(ttygets(file_desc,cap,capstr,20) == 0) capstr[0] = '\0';
   return(capstr);
}

/***************************************************************/

static int
read_fits_header(record,nx,ny,dtype,dsize,bscale,bbzero,xmin,xmax,ymin,ymax)
char *record;
int *nx,*ny,                            /* size of file */
    *dtype,                             /* type of data */
    *dsize;                             /* length of data */
float *bscale,                          /* data = bbzero + I*bscale */
      *bbzero,
      *xmin,*xmax,                      /* range of values on the x axis */
      *ymin,*ymax;                      /* and on the y axis */
{
   char buff[81],                       /* for reading value (81 plays safe) */
	*keyword,                       /* keyword from a FITS card */
	*value;                         /* value of a FITS keyword */
   int i,
       naxis = 0;                       /* number of axes in data */

   i = 0;
   if(*nx == 0) {                       /* first record of header */
      if(read_card(record,&keyword,&value) < 0) {
	 return(-1);
      }
      if(strcmp(keyword,"SIMPLE") != 0) {
	 msg("First keyword on FITS header isn't SIMPLE\n");
	 return(-1);
      }
      (void)str_scanf(value,"%s",buff);
      if(!strcmp(buff,"T")) {           /* OK, simple FITS. Good */
	 *dtype = INT;                  /* for now, pending value of BITPIX */
      } else if(!strcmp(buff,"F")) {    /* Assume it is floating */
	 if(sm_verbose) {
	    msg("FITS file isn't simple, assuming `float'\n");
	 }
	 *dtype = FLOAT;
      } else {
	 msg_1s("Illegal value for keyword SIMPLE : %s\n",value);
	 return(-1);
      }
      i++; record += 80;

      if(read_card(record,&keyword,&value) < 0) {
	 return(-1);
      }
      if(strcmp(keyword,"BITPIX") != 0) {
	 msg("Second keyword on FITS header isn't BITPIX\n");
	 return(-1);
      }
      *dsize = atoi(value)/BITS_IN_CHAR;
      if(*dsize < 0) {			/* floating IEEE */
	 *dsize = -(*dsize);
	 *dtype = FLOAT;
      }
      if(*dtype == FLOAT) {
	 if(*dsize != 4) {
	    msg_1d("Illegal value of BITPIX for floating data: %d\n",
								atoi(value));
	    return(-1);
	 }
      } else if(*dsize == 1) {
	 *dtype = CHAR;
      } else if(*dsize == 2) {
	 *dtype = SHORT;
      } else if(*dsize == 4) {
	 *dtype = LONG;
      } else {
	 msg_1d("Illegal value of BITPIX: %d\n",atoi(value));
	 return(-1);
      }
      i++; record += 80;

      if(read_card(record,&keyword,&value) < 0) {
	 return(-1);
      }
      if(strcmp(keyword,"NAXIS") != 0) {
	 msg("Third keyword on FITS header isn't NAXIS\n");
	 return(-1);
      }
      naxis = atoi(value);
      i++; record += 80;

      if(read_card(record,&keyword,&value) < 0) {
	 return(-1);
      }
      if(strcmp(keyword,"NAXIS1") != 0) {
	 msg("Fourth keyword on FITS header isn't NAXIS1\n");
	 return(-1);
      }
      *nx = atoi(value);
      i++; record += 80;

      if(naxis == 1) {
	 *ny = 1;
      } else {
	 if(read_card(record,&keyword,&value) < 0) {
	    return(-1);
	 }
	 if(strcmp(keyword,"NAXIS2") != 0) {
	    msg("Fifth keyword on FITS header isn't NAXIS2\n");
	    return(-1);
	 }
	 *ny = atoi(value);
	 i++; record += 80;
      }
   }
/*
 * We've read the compulsary header, now look for END
 */
   for(;i < 36;i++) {
      if(read_card(record,&keyword,&value) < 0) {
	 continue;			/* try to recover */
      } else if(!strcmp(keyword,"END")) {
	 break;
      }
      record += 80;
   }

   if(naxis > 2) {
      char buff[10];

      for(i = 3;i < naxis;i++) {
	 sprintf(buff,"NAXIS%d",i);
	 if(*(value = get_keyword(buff)) != '\0' && atoi(value) != 1) {
	    msg_1d("NAXIS isn't 1 or 2 (it's %d), ",naxis);
	    msg_1d("NAXIS%d = ",i);
	    msg_1d("%d, so data isn't 2-dimensional.\n",atoi(value));
	    msg_2s("%s%s\n",
		"I'll ignore the extra dimensions, ",
		"and pretend that NAXIS is 2.");
	 }
      }
   }
   if(*(value = get_keyword("BSCALE")) != '\0') *bscale = atof2(value);
   if(*(value = get_keyword("BZERO")) != '\0') *bbzero = atof2(value);

   *xmin = 0;
   *xmax = *nx - 1;
   if(*get_keyword("X0") != '\0' && *get_keyword("X1") != '\0') {
      *xmin = atof2(get_keyword("X0"));
      *xmax = atof2(get_keyword("X1"));
   } else {
      float crval,crpix,cdelt;
      char *val1,*val2,*val3;
      
      val1 = get_keyword("CRPIX1");
      val2 = get_keyword("CRVAL1");
      val3 = get_keyword("CDELT1");

      if(*val1 != '\0' && *val2 != '\0' && *val3 != '\0') {
	 if(*(value = get_keyword("CROTA1")) != '\0' && atof2(value) != 0) {
	    msg_1f("CROTA1 = %f in non-zero; I'm it ignoring\n",atof2(value));
	 } else {
	    crpix = atof2(val1);
	    crval = atof2(val2);
	    cdelt = atof2(val3);
	    *xmin = crval + (1 - crpix)*cdelt;
	    *xmax = crval + (*nx - crpix)*cdelt;
	    (void)sprintf(buff,"%g",*xmin); set_image_variable("X0",buff);
	    (void)sprintf(buff,"%g",*xmax); set_image_variable("X1",buff);
	 }
      }
   }

   *ymin = 0;
   *ymax = *ny - 1;
   if(*get_keyword("Y0") != '\0' && *get_keyword("Y1") != '\0') {
      *ymin = atof2(get_keyword("Y0"));
      *ymax = atof2(get_keyword("Y1"));
   } else {
      float crval,crpix,cdelt;
      char *val1,*val2,*val3;
      
      val1 = get_keyword("CRPIX2");
      val2 = get_keyword("CRVAL2");
      val3 = get_keyword("CDELT2");
      if(*val1 != '\0' && *val2 != '\0' && *val3 != '\0') {
	 if(*(value = get_keyword("CROTA2")) != '\0' && atof2(value) != 0) {
	    msg_1f("CROTA2 = %f in non-zero; I'm it ignoring\n",atof2(value));
	 } else {
	    crpix = atof2(val1);
	    crval = atof2(val2);
	    cdelt = atof2(val3);
	    *ymin = crval + (1 - crpix)*cdelt;
	    *ymax = crval + (*ny - crpix)*cdelt;
	    (void)sprintf(buff,"%g",*ymin); set_image_variable("Y0",buff);
	    (void)sprintf(buff,"%g",*ymax); set_image_variable("Y1",buff);
	 }
      }
   }

   return(i == 36 ? 1 : 0);
}

static int
read_card(card,keyword,value)
char *card,
     **keyword,
     **value;
{
   int i;

   *keyword = &card[0];

   if(!(!strncmp(*keyword,"COMMENT",7) || !strncmp(*keyword,"END",3) ||
	!strncmp(*keyword,"HISTORY",7) || !strncmp(*keyword,"        ",8))) {
      if(card[8] != '=' || card[9] != ' ') {
	 (*keyword)[8] = '\0';
	 if(sm_verbose) {
	    msg_1s("Missing = for keyword %s\n",*keyword);
	 }
      }
   }

   (*keyword)[8] = '\0';
   for(i = 7;i >= 0;i--) {
      if((*keyword)[i] == ' ') (*keyword)[i] = '\0';
   }
   if((*keyword)[0] == '\0') {
      msg("missing keyword in FITS header\n");
      return(-1);
   }

   for(i = 10;isspace(card[i]);i++) ;	/* skip leading space */
   if(card[i] == '\'') {		/* character value */
      *value = &card[++i];
      while(i < 80 && card[i] != '\'') i++;
   } else {				/* non-character value */
      *value = &card[i];
      while(i < 80 && !isspace(card[i]) && card[i] != '/') i++;
   }
   card[i] = '\0';

   set_image_variable(*keyword,*value);
   return(0);
}

/*
 * Byte swap ABABAB -> BABABAB in place
 */
static void
swap_2(arr,n)
char *arr;                              /* array to swap */
int n;                                  /* number of bytes */
{
   char *end,
	t;

   if(n%2 != 0) {
      msg_1d("Attempt to byte swap odd number of bytes: %d\n",n);
      n = 2*(int)(n/2);
   }

   for(end = arr + n;arr < end;arr += 2) {
      t = arr[0];
      arr[0] = arr[1];
      arr[1] = t;
   }
}

/*
 * Byte swap ABCDABCD -> DCBADCBA in place (e.g. sun <--> vax)
 */
static void
swap_4(arr,n)
char *arr;                              /* array to swap */
int n;                                  /* number of bytes */
{
   char *end,
	t;

   if(n%4 != 0) {
      msg_1d("Attempt to byte swap non-multiple of 4 bytes: %d\n",n);
      n = 4*(int)(n/4);
   }

   for(end = arr + n;arr < end;arr += 4) {
      t = arr[0];
      arr[0] = arr[3];
      arr[3] = t;
      t = arr[1];
      arr[1] = arr[2];
      arr[2] = t;
   }
}

/***************************************************************/
/*
 * Stuff to handle the tree for keywords
 */
#define VALSIZE 40
typedef struct {
   char value[VALSIZE + 1];
} KNODE;

static char *kvalue;
static void kkill();
static Void *kmake();
static TREE kk = { NULL, kkill, kmake};

/*
 * Add a keyword to the tree
 */
void
set_image_variable(name,value)
char *name;
char *value;
{
   kvalue = value;

   insert_node(name,&kk);
}

/*
 * return value of keyword
 */
static char *
get_keyword(name)
char *name;
{
   KNODE *node;

   if((node = (KNODE *)get_rest(name,&kk)) == NULL) {/* unknown keyword */
      return("");
   } else {
      return(node->value);
   }
}

/*
 * Forget all keywords
 */
void
kill_keywords()
{
   kill_tree(&kk);
}

static void
kkill(nn)
Void *nn;
{
   free((char *)nn);
}

/*
 * make new node
 */
static Void *
kmake(name,nn)
char *name;
Void *nn;			/* current value of node */
{
   KNODE *node;
   
   if(nn == NULL) {
      if((node = (KNODE *)malloc(sizeof(KNODE))) == NULL) {
	 msg_1s("malloc returns NULL in vmake() for %s\n",name);
	 return(NULL);
      }
      node->value[VALSIZE] = '\0';		/* Check value is terminated */
   } else {
      node = (KNODE *)nn;
   }
   (void)strncpy(node->value,kvalue,VALSIZE);

   return((Void *)node);
}
