#include "copyright.h"
/*
 * This file contains a number of functions to do arithmetic and string
 * operations on vectors.
 *
 *    function			finds
 *
 * vec_abs(v1,ans)		ans = abs(v1)
 * vec_acos(v1,ans)		ans = acos(v1)
 * vec_add(v1,v2,ans)		ans = v1 + v2
 * vec_asin(v1,ans)		ans = asin(v1)
 * vec_atan(v1,ans)		ans = atan(v1)
 * vec_concat(v1,v2,ans)	ans = concatenation of v1 and v2
 * vec_cos(v1,ans)		ans = cos(v1)
 * vec_divide(v1,v2,ans)	ans = v1 / v2
 * vec_hist(v1,v2,ans)		ans = histogram of v1 at v2
 * vec_interp(v1,v2,ans)	ans = IMAGE(v1, v2)
 * vec_int(v1,ans)		ans = (int)v1
 * vec_lg(v1,ans)		ans = lg(v1)
 * vec_ln(v1,ans)		ans = ln(v1)
 * vec_multiply(v1,v2,ans)	ans = v1 * v2
 * vec_power(v1,v2,ans)		ans = v1^v2
 * vec_random(v1,ans)		ans = random number with same dimension as v1
 * vec_set_image(v1,v2,vals)    IMAGE(v1,v2) = vals
 * vec_sin(v1,ans)		ans = sin(v1)
 * vec_sqrt(v1,ans)		ans = sqrt(v1)
 * vec_subtract(v1,v2,ans)	ans = v1 - v2
 * vec_tan(v1,ans)		ans = tan(v1)
 */
#include <stdio.h>
#include <math.h>
#include "options.h"
#include "vectors.h"
#include "sm_declare.h"

extern int sm_interrupt,			/* Have we seen a ^C? */
  	   sm_verbose;			/* shall we be chatty? */

/***********************************************************/
/*
 * Take the absolute value of a vector
 */
void
vec_abs(v1,ans)
VECTOR *v1,			/* vector to be absed */
       *ans;			/* the answer */
{
   int i;

   if(check_vec(v1,(VECTOR *)NULL,ans) < 0) return;
   
   for(i = 0;i < v1->dimen;i++) {
      v1->vec.f[i] = fabs(v1->vec.f[i]);
   }
   ans->dimen = v1->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v1->vec.f;
   return;
}

/***********************************************************/
/*
 * Take the acos of a vector
 */
void
vec_acos(v1,ans)
VECTOR *v1,			/* vector to be acosed */
       *ans;			/* the answer */
{
   int i,
       first_err = 1;			/* First arithmetic error? */

   if(check_vec(v1,(VECTOR *)NULL,ans) < 0) return;
   
   for(i = 0;i < v1->dimen;i++) {
      if(fabs(v1->vec.f[i]) <= 1) {
	 v1->vec.f[i] = acos(v1->vec.f[i]);
      } else {
	 if(sm_verbose) {
	    if(sm_verbose > 1 || first_err) {
	       first_err = 0;
	       msg_1f("Attempt to take acos of %g",v1->vec.f[i]);
	       yyerror("show where");
	    }
	 }
	 if(sm_interrupt) break;
	 v1->vec.f[i] = 0;
      }
   }
   ans->dimen = v1->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v1->vec.f;
   return;
}

/********************************************************/
/*
 * Add two vectors
 */
void
vec_add(v1,v2,ans)
VECTOR *v1,*v2,			/* vectors to be added */
       *ans;			/* the answer */
{
   char buff[2*VEC_STR_SIZE];
   int i,
       exchange = 0,		/* have vectors been exchanged ? */
       first_err = 1;			/* First error? */
   VECTOR *temp;		/* used in switching v1 and v2 */

   if(v1->type != v2->type) {
      msg_2s("Vectors %s and %s have different types",v1->name,v2->name);
      yyerror("show where");
      vec_free(v1);
      vec_free(v2);
      if(ans != NULL) {
	 ans->type = VEC_NULL;
	 ans->dimen = 0;
	 ans->name = "(empty)";
      }
      return;
   }
   
   if(v1->dimen < v2->dimen) {	/* make v1 >= v2 */
      temp = v1;
      v1 = v2;
      v2 = temp;
      exchange = 1;
   }

   if(v2->dimen == 0) {			/* swap them back, to free longer */
      temp = v1;
      v1 = v2;
      v2 = temp;
      exchange = !exchange;
   } else if(v2->dimen == 1) {		/* add scalar and vector */
      if(v1->type == VEC_FLOAT) {
	 for(i = 0;i < v1->dimen;i++) {
	    v1->vec.f[i] += v2->vec.f[0];
	 }
      } else {
	 if(exchange) {
	    for(i = 0;i < v1->dimen;i++) {
	       (void)sprintf(buff,"%s%s",v2->vec.s.s[0],v1->vec.s.s[i]);
	       (void)strncpy(v1->vec.s.s[i],buff,VEC_STR_SIZE-1);
	       if(strlen(buff) >= VEC_STR_SIZE) {
		  if(sm_verbose > 1 || first_err) {
		     first_err = 0;
		     msg_1d("Truncating %dth element of string vector",i);
		     msg_2s(" \"%s + %s\"",v2->name,v1->name);
		     yyerror("show where");
		  }
		  v1->vec.s.s[i][VEC_STR_SIZE-1] = '\0';
	       }
	    }
	 } else {
	    for(i = 0;i < v1->dimen;i++) {
	       (void)sprintf(buff,"%s%s",v1->vec.s.s[i],v2->vec.s.s[0]);
	       (void)strncpy(v1->vec.s.s[i],buff,VEC_STR_SIZE-1);
	       if(strlen(buff) >= VEC_STR_SIZE) {
		  if(sm_verbose > 1 || first_err) {
		     first_err = 0;
		     msg_1d("Truncating %dth element of string vector",i);
		     msg_2s(" \"%s + %s\"",v1->name,v2->name);
		     yyerror("show where");
		  }
		  v1->vec.s.s[i][VEC_STR_SIZE-1] = '\0';
	       }
	    }
	 }
      }
   } else {			/* both vectors */
      if(v1->dimen != v2->dimen) {
	 msg_2s("%s + %s: vectors' dimensions differ",v1->name,v2->name);
	 yyerror("show where");
      }
      if(v1->type == VEC_FLOAT) {
	 for(i = 0;i < v2->dimen;i++) {
	    v1->vec.f[i] += v2->vec.f[i];
	 }
      } else {
	 if(exchange) {
	    for(i = 0;i < v2->dimen;i++) {
	       (void)sprintf(buff,"%s%s",v2->vec.s.s[i],v1->vec.s.s[i]);
	       (void)strncpy(v1->vec.s.s[i],buff,VEC_STR_SIZE-1);
	       if(strlen(buff) >= VEC_STR_SIZE) {
		  if(sm_verbose > 1 || first_err) {
		     first_err = 0;
		     msg_1d("Truncating %dth element of string vector",i);
		     msg_2s(" \"%s + %s\"",v2->name,v1->name);
		     yyerror("show where");
		  }
		  v1->vec.s.s[i][VEC_STR_SIZE-1] = '\0';
	       }
	    }
	 } else {
	    for(i = 0;i < v2->dimen;i++) {
	       (void)sprintf(buff,"%s%s",v1->vec.s.s[i],v2->vec.s.s[i]);
	       (void)strncpy(v1->vec.s.s[i],buff,VEC_STR_SIZE-1);
	       if(strlen(buff) >= VEC_STR_SIZE) {
		  if(sm_verbose > 1 || first_err) {
		     first_err = 0;
		     msg_1d("Truncating %dth element of string vector",i);
		     msg_2s(" \"%s + %s\"",v1->name,v2->name);
		     yyerror("show where");
		  }
		  v1->vec.s.s[i][VEC_STR_SIZE-1] = '\0';
	       }
	    }
	 }
      }
   }
   vec_free(v2);
   ans->dimen = v1->dimen;
   ans->type = v1->type;
   ans->name = v1->name;
   ans->vec = v1->vec;
   return;
}

/***********************************************************/
/*
 * Take the asin of a vector
 */
void
vec_asin(v1,ans)
VECTOR *v1,			/* vector to be asined */
       *ans;			/* the answer */
{
   int i,
       first_err = 1;			/* First arithmetic error? */

   if(check_vec(v1,(VECTOR *)NULL,ans) < 0) return;

   for(i = 0;i < v1->dimen;i++) {
      if(fabs(v1->vec.f[i]) <= 1) {
	 v1->vec.f[i] = asin(v1->vec.f[i]);
      } else {
	 if(sm_verbose) {
	    if(sm_verbose > 1 || first_err) {
	       first_err = 0;
	       msg_1f("Attempt to take asin of %g",v1->vec.f[i]);
	       yyerror("show where");
	    }
	 }
	 if(sm_interrupt) break;
	 v1->vec.f[i] = 0;
      }
   }
   ans->dimen = v1->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v1->vec.f;
   return;
}

/***********************************************************/
/*
 * Take the atan of a vector
 */
void
vec_atan(v1,ans)
VECTOR *v1,			/* vector to be ataned */
       *ans;			/* the answer */
{
   int i;

   if(check_vec(v1,(VECTOR *)NULL,ans) < 0) return;

   for(i = 0;i < v1->dimen;i++) {
      v1->vec.f[i] = atan(v1->vec.f[i]);
   }
   ans->dimen = v1->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v1->vec.f;
   return;
}

/***********************************************************/
/*
 * Take the atan of a pair of vectors, x and y
 */
void
vec_atan2(v1,v2,ans)
VECTOR *v1,*v2,				/* vectors to be atan2ed */
       *ans;				/* the answer */
{
   int i,
       exchange = 0;			/* have vectors been exchanged ? */
   VECTOR *temp;			/* used in exchangeing v1 and v2 */

   if(check_vec(v1,v2,ans) < 0) return;

   if(v1->dimen < v2->dimen) {	/* make v1 >= v2 */
      temp = v1;
      v1 = v2;
      v2 = temp;
      exchange = 1;
   }

   if(v2->dimen == 0) {			/* swap them back, to free longer */
      temp = v1;
      v1 = v2;
      v2 = temp;
   } else if(v2->dimen == 1) {		/* scalar and vector */
      if(!exchange) {
	 for(i = 0;i < v1->dimen;i++) {
	    v1->vec.f[i] = atn2(v1->vec.f[i],v2->vec.f[0]);
	 }
      } else {
	 for(i = 0;i < v1->dimen;i++) {
	    v1->vec.f[i] = atn2(v2->vec.f[0],v1->vec.f[i]);
	 }
      }
   } else {			/* both vectors */
      if(v1->dimen != v2->dimen) {
	 msg_2s("atan2(%s,%s): vectors' dimensions differ",v1->name,v2->name);
	 yyerror("show where");
      }
      for(i = 0;i < v2->dimen;i++) {
	 if(!exchange) {
	    v1->vec.f[i] = atn2(v1->vec.f[i],v2->vec.f[i]);
	 } else {
	    v1->vec.f[i] = atn2(v2->vec.f[i],v1->vec.f[i]);
	 }
      }
   }
   vec_free(v2);
   ans->dimen = v1->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v1->vec.f;
   return;
}

/********************************************************/
/*
 * Concatenate two vectors
 */
void
vec_concat(v1,v2,ans)
VECTOR *v1,*v2,			/* vectors to be added */
       *ans;			/* the answer */
{
   int i,
       old_v1_dimen;			/* dimension of v1 before reallocing */

   if(v1->type != v2->type) {
      msg_2s("Vectors %s and %s have different types in CONCAT",
	     v1->name,v2->name);
      yyerror("show where");
      vec_free(v1);
      vec_free(v2);
      (void)vec_value(ans,0.0);
      return;
   }

   old_v1_dimen = v1->dimen;
   if(vec_realloc(v1,v1->dimen + v2->dimen) < 0) {
      msg("realloc returns NULL in vec_concat()\n");
      ans->dimen = 0;
      ans->type = VEC_NULL;
      ans->name = "(failed concat)";
      return;
   }
   ans->vec = v1->vec;
   ans->dimen = v1->dimen;
   ans->type = v1->type;
   ans->name = "(concated)";
   if(ans->type == VEC_FLOAT) {
      for(i = 0;i < v2->dimen;i++) {
	 ans->vec.f[i + old_v1_dimen] = v2->vec.f[i];
      }
   } else if(ans->type == VEC_STRING) {
      for(i = 0;i < v2->dimen;i++) {
	 (void)strncpy(ans->vec.s.s[i + old_v1_dimen],v2->vec.s.s[i],
		       VEC_STR_SIZE);
      }
   } else {
      msg_2s("Vectors %s and %s have ",v1->name,v2->name);
      msg_1d("unknown type %d in CONCAT\n",ans->type);
      vec_free(v1);
      (void)vec_value(ans,0.0);
   }
   vec_free(v2);
   return;
}

/***********************************************************/
/*
 * Take the cosine of a vector
 */ 
void
vec_cos(v1,ans)
VECTOR *v1,			/* vector to be cosined */
       *ans;			/* the answer */
{
   int i;

   if(check_vec(v1,(VECTOR *)NULL,ans) < 0) return;

   for(i = 0;i < v1->dimen;i++) {
      v1->vec.f[i] = cos(v1->vec.f[i]);
   }
   ans->dimen = v1->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v1->vec.f;
   return;
}

/********************************************************/
/*
 * divide two vectors
 */
void
vec_divide(v1,v2,ans)
VECTOR *v1,*v2,			/* vectors to be divided */
       *ans;			/* the answer */
{
   int i,
       exchange = 0,		/* have vectors been exchanged ? */
       first_err = 1;			/* First arithmetic error? */
   VECTOR *temp;		/* used in exchangeing v1 and v2 */

   if(check_vec(v1,v2,ans) < 0) return;

   if(v1->dimen < v2->dimen) {	/* make v1 >= v2 */
      temp = v1;
      v1 = v2;
      v2 = temp;
      exchange = 1;
   }

   if(v2->dimen == 0) {			/* swap them back, to free longer */
      temp = v1;
      v1 = v2;
      v2 = temp;
   } else if(v2->dimen == 1) {		/* scalar and vector */
      if(exchange) {
	 for(i = 0;i < v1->dimen;i++) {
	    if(v1->vec.f[i] == 0) {
	       if(sm_verbose) {
		  if(sm_verbose > 1 || first_err) {
		     first_err = 0;
		     msg("Attempt to divide by zero");
		     yyerror("show where");
		  }
	       }
	       if(sm_interrupt) break;
	       v1->vec.f[i] = 0;
	    } else {
	       v1->vec.f[i] = v2->vec.f[0]/v1->vec.f[i];
	    }
	 }
      } else {
	 if(v2->vec.f[0] == 0) {
	    if(sm_verbose) {
	       if(sm_verbose > 1 || first_err) {
		  first_err = 0;
		  msg("Attempt to divide by zero");
		  yyerror("show where");
	       }
	    }
	    for(i = 0;i < v1->dimen;i++) {
	       v1->vec.f[i] = 0;
	    }
	 } else {
	    for(i = 0;i < v1->dimen;i++) {
	       v1->vec.f[i] /= v2->vec.f[0];
	    }
	 }
      }
   } else {			/* both vectors */
      if(v1->dimen != v2->dimen) {
	 msg_2s("%s / %s: vectors' dimensions differ",v1->name,v2->name);
	 yyerror("show where");
      }
      for(i = 0;i < v2->dimen;i++) {
	 if(!exchange) {
	    if(v2->vec.f[i] != 0) {
	       v1->vec.f[i] /= v2->vec.f[i];
	    } else {
	       if(sm_verbose) {
		  if(sm_verbose > 1 || first_err) {
		     first_err = 0;
		     msg("Attempt to divide by zero");
		     yyerror("show where");
		  }
	       }
	       if(sm_interrupt) break;
	       v1->vec.f[i] = 0;
	    }
	 } else {
	    if(v1->vec.f[i] != 0) {
	       v1->vec.f[i] = v2->vec.f[i]/v1->vec.f[i];
	    } else {
	       if(sm_verbose) {
		  if(sm_verbose > 1 || first_err) {
		     first_err = 0;
		     msg("Attempt to divide by zero");
		     yyerror("show where");
		  }
	       }
	       if(sm_interrupt) break;
	       v1->vec.f[i] = 0;
	    }
	 }
      }
   }
   vec_free(v2);
   ans->dimen = v1->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v1->vec.f;
   return;
}

/***********************************************************/
/*
 * Take the exponential of a vector
 */
void
vec_exp(v1,ans)
VECTOR *v1,			/* vector to be exponentiated */
       *ans;			/* the answer */
{
   int i;

   if(check_vec(v1,(VECTOR *)NULL,ans) < 0) return;

   for(i = 0;i < v1->dimen;i++) {
      v1->vec.f[i] = exp(v1->vec.f[i]);
   }
   ans->dimen = v1->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v1->vec.f;
   return;
}

/***********************************************************/
/*
 * Convert v1 into a histogram at points v2. That is, return a histogram
 * where ans[i] is the number of elements in v1 lying in
 * [ (v2[i-1]+v2[i])/2 , (v2[i]+v2[i+1])/2 ]
 */
void
vec_hist(v1,v2,ans)
VECTOR *v1,*v2,			/* vectors giving coordinates */
       *ans;			/* the answer */
{
   float mid_val;			/* value at middle of a bin */
   int i,j,k;

   if(check_vec(v1,v2,ans) < 0) return;

   for(i = 1;i < v2->dimen;i++) {
      if(v2->vec.f[i] < v2->vec.f[i-1]) {
         msg("Second vector is not sorted in vec_hist\n");
	 ans = v1;
	 vec_free(v2);
	 return;
      }
   }

   sort_flt(v1->vec.f,v1->dimen);

   i = k = 0;
   for(j = 0;j < v2->dimen - 1;j++) {
      mid_val = (v2->vec.f[j] + v2->vec.f[j+1])/2;	/* midpoint of bin */
      for(;i < v1->dimen;i++) {
         if(v1->vec.f[i] > mid_val) {
            v2->vec.f[j] = i - k;
	    k = i;
	    break;
	 }
      }
      if(i == v1->dimen) break;
   }
   v2->vec.f[j++] = v1->dimen - k;
   for(;j < v2->dimen;j++) {
      v2->vec.f[j] = 0;
   }
	    
   vec_free(v1);
   ans->dimen = v2->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v2->vec.f;
   return;
}

/***********************************************************/
/*
 * Interpolate into the current IMAGE at points (v1,v2)
 */
void
vec_interp(v1,v2,ans)
VECTOR *v1,*v2,			/* vectors giving coordinates */
       *ans;			/* the answer */
{
   int exchange = 0,		/* did I switch v1 and v2 ? */
       i;
   VECTOR *temp;		/* used in switching v1 and v2 */

   if(check_vec(v1,(VECTOR *)NULL,ans) < 0) return;

   if(v1->dimen < v2->dimen) {	/* make v1 >= v2 */
      temp = v1;
      v1 = v2;
      v2 = temp;
      exchange = 1;
   }

   if(v2->dimen == 1) {		/* scalar and vector */
      if(!exchange) {
         for(i = 0;i < v1->dimen;i++) {
	    v1->vec.f[i] = image_val(v1->vec.f[i],v2->vec.f[0]);
         }
      } else {
         for(i = 0;i < v1->dimen;i++) {
	    v1->vec.f[i] = image_val(v2->vec.f[0],v1->vec.f[i]);
         }
      }
   } else {			/* both vectors */
      if(v1->dimen != v2->dimen) {
	 msg("X and Y vectors are different lengths");
	 yyerror("show where");
      }
      for(i = 0;i < v2->dimen;i++) {
	 v1->vec.f[i] = image_val(v1->vec.f[i],v2->vec.f[i]);
      }
   }
   vec_free(v2);
   ans->dimen = v1->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v1->vec.f;
   return;
}

/***********************************************************/
/*
 * Take the integral part of a vector
 */
void
vec_int(v1,ans)
VECTOR *v1,			/* vector to be cast to int */
       *ans;			/* the answer */
{
   int i;

   if(check_vec(v1,(VECTOR *)NULL,ans) < 0) return;

   for(i = 0;i < v1->dimen;i++) {
      v1->vec.f[i] = (int)(v1->vec.f[i]);
   }
   ans->dimen = v1->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v1->vec.f;
   return;
}

/***********************************************************/
/*
 * Take the lg of a vector
 */
void
vec_lg(v1,ans)
VECTOR *v1,			/* vector to be logged (base 10) */
       *ans;			/* the answer */
{
   int i,
       first_err = 1;			/* First arithmetic error? */

   if(check_vec(v1,(VECTOR *)NULL,ans) < 0) return;

   for(i = 0;i < v1->dimen;i++) {
      if(v1->vec.f[i] > 0) {
	 v1->vec.f[i] = log10(v1->vec.f[i]);
      } else {
	 if(sm_verbose) {
	    if(sm_verbose > 1 || first_err) {
	       first_err = 0;
	       msg("Attempt to take lg of non-positive number");
	       yyerror("show where");
	    }
	 }
	 v1->vec.f[i] = -1000;
	 if(sm_interrupt) break;
      }
   }
   ans->dimen = v1->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v1->vec.f;
   return;
}

/***********************************************************/
/*
 * Take the lg of a vector
 */
void
vec_ln(v1,ans)
VECTOR *v1,			/* vector to be natural logged */
       *ans;			/* the answer */
{
   int i,
       first_err = 1;			/* First arithmetic error? */

   if(check_vec(v1,(VECTOR *)NULL,ans) < 0) return;

   for(i = 0;i < v1->dimen;i++) {
      if(v1->vec.f[i] > 0) {
	 v1->vec.f[i] = log(v1->vec.f[i]);
      } else {
	 if(sm_verbose) {
	    if(sm_verbose > 1 || first_err) {
	       first_err = 0;
	       msg("Attempt to take ln of non-positive number");
	       yyerror("show where");
	    }
	 }
	 v1->vec.f[i] = -1000;
	 if(sm_interrupt) break;
      }
   }
   ans->dimen = v1->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v1->vec.f;
   return;
}

/***********************************************************/
/*
 * Multiply two vectors
 */
void
vec_multiply(v1,v2,ans)
VECTOR *v1,*v2,			/* vectors to be added */
       *ans;			/* the answer */
{
   int i;
   VECTOR *temp;		/* used in switching v1 and v2 */

   if(check_vec(v1,v2,ans) < 0) return;

   if(v1->dimen < v2->dimen) {	/* make v1 >= v2 */
      temp = v1;
      v1 = v2;
      v2 = temp;
   }

   if(v2->dimen == 0) {			/* swap them back, to free longer */
      temp = v1;
      v1 = v2;
      v2 = temp;
   } else if(v2->dimen == 1) {		/* scalar and vector */
      for(i = 0;i < v1->dimen;i++) {
	 v1->vec.f[i] *= v2->vec.f[0];
      }
   } else {			/* both vectors */
      if(v1->dimen != v2->dimen) {
	 msg_2s("%s * %s: vectors' dimensions differ",v1->name,v2->name);
	 yyerror("show where");
      }
      for(i = 0;i < v2->dimen;i++) {
	 v1->vec.f[i] *= v2->vec.f[i];
      }
   }
   vec_free(v2);
   ans->dimen = v1->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v1->vec.f;
   return;
}

/***********************************************************/
/*
 * Calculate a power of a vector
 */
void
vec_power(v1,v2,ans)
VECTOR *v1,			/* vector to be exponentiated */
       *v2,			/* indexes to be used */
       *ans;			/* the answer */
{
   float base,			/* we want base^index */
	 index;
   int exchange = 0,		/* have vectors been exchanged ? */
       first_err = 1,			/* First arithmetic error? */
       i,
       integral;		/* is index integral ? */
   VECTOR *temp;		/* used in exchangeing v1 and v2 */

   if(check_vec(v1,v2,ans) < 0) return;

   if(v1->dimen < v2->dimen) {	/* make v1 >= v2 */
      temp = v1;
      v1 = v2;
      v2 = temp;
      exchange = 1;
   }

   if(v2->dimen == 0) {			/* swap them back, to free longer */
      temp = v1;
      v1 = v2;
      v2 = temp;
   } else if(v2->dimen == 1) {		/* scalar and vector */
      if(exchange) {
	 base = v2->vec.f[0];
	 for(i = 0;i < v1->dimen;i++) {
	    index = v1->vec.f[i];
	    if(base < 0 && fabs(index) - (int)fabs(index) != 0) {
	       if(sm_verbose) {
		  if(sm_verbose > 1 || first_err) {
		     first_err = 0;
		     msg("Attempt to take fractional power of -ve number");
		     yyerror("show where");
		  }
	       }
	       if(sm_interrupt) break;
	       v1->vec.f[i] = 0;
	    } else {
	       v1->vec.f[i] = pow(base,index);
	    }
	 }
      } else {
	 index = v2->vec.f[0];
	 integral = (fabs(index) - (int)fabs(index) == 0) ? 1 : 0;
	 for(i = 0;i < v1->dimen;i++) {
	    base = v1->vec.f[i];
	    if(!integral && base < 0) {
	       if(sm_verbose) {
		  if(sm_verbose > 1 || first_err) {
		     first_err = 0;
		     msg("Attempt to take fractional power of -ve number");
		     yyerror("show where");
		  }
	       }
	       if(sm_interrupt) break;
	       v1->vec.f[i] = 0;
	    } else {
	       v1->vec.f[i] = pow(base,index);
	    }
	 }
      }
   } else {			/* both vectors */
      if(v1->dimen != v2->dimen) {
	 msg_2s("%s ** %s: vectors' dimensions differ",v1->name,v2->name);
	 yyerror("show where");
      }
      for(i = 0;i < v2->dimen;i++) {
	 if(exchange) {
	    base = v2->vec.f[i];
	    index = v1->vec.f[i];
	 } else {
	    index = v2->vec.f[i];
	    base = v1->vec.f[i];
	 }
	 if(base < 0 && fabs(index) - (int)fabs(index) != 0) {
	    if(sm_verbose) {
	       if(sm_verbose > 1 || first_err) {
		  first_err = 0;
		  msg("Attempt to take fractional power of -ve number");
		  yyerror("show where");
	       }
	    }
	    if(sm_interrupt) break;
	    v1->vec.f[i] = 0;
	 } else {
	    v1->vec.f[i] = pow(base,index);
	 }
      }
   }
   vec_free(v2);
   ans->dimen = v1->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v1->vec.f;
   return;
}

/***********************************************************/
/*
 * Return a vector of random numbers in [0,1]
 */
static int init_random = 0;		/* is the random number generator
					   initialised? */

void
vec_random(size,vals)
int size;
VECTOR *vals;				/* the values */
{
   int i;
   
   if(make_anon_vector(vals,size,VEC_FLOAT) == -1) {
      return;
   }

   if(!init_random) {
      SET_RAND(time((long *)NULL));
      init_random = 1;
   }

   for(i = 0;i < size;i++) {
      vals->vec.f[i] = RAND()/(double)MAX_RAND;
   }
}

void
set_random(seed)
long seed;
{
   SET_RAND(seed);
   init_random = 1;
}

/***********************************************************/
/*
 * Set the value of the current IMAGE(v1,v2) = v3
 */
void
vec_set_image(v1,v2,vals)
VECTOR *v1,*v2,				/* vectors giving coordinates */
       *vals;				/* the values */
{
   int exchange = 0,			/* did I switch v1 and v2 ? */
       first_err = 1,			/* First arithmetic error? */
       i,n;
   VECTOR *temp;			/* used in switching v1 and v2 */

   if(v1->type != VEC_FLOAT ||
      v2->type != VEC_FLOAT || vals->type != VEC_FLOAT) {
      if(v1->type != VEC_FLOAT) {
	 msg_1s("Vector %s is not arithmetic",v1->name);
	 yyerror("show where");
      }
      if(v2->type != VEC_FLOAT) {
	 msg_1s("Vector %s is not arithmetic",v2->name);
	 yyerror("show where");
      }
      if(vals->type != VEC_FLOAT) {
	 msg_1s("Vector %s is not arithmetic",vals->name);
	 yyerror("show where");
      }
      vec_free(v1);
      vec_free(v2);
      vec_free(vals);
      return;
   }

   if(v1->dimen < v2->dimen) {		/* make v1 >= v2 */
      temp = v1;
      v1 = v2;
      v2 = temp;
      exchange = 1;
   }

   if(v2->dimen == 1) {			/* scalar and vector */
      if(vals->dimen == 1) {
	 if(!exchange) {
	    for(i = 0;i < v1->dimen;i++) {
	       if(set_image_val((int)v1->vec.f[i],(int)v2->vec.f[0],
			     vals->vec.f[0]) < 0) {
		  if(sm_verbose > 1 || first_err) {
		     first_err = 0;
		     msg_1d("IMAGE(%d,",(int)v1->vec.f[i]);
		     msg_1d("%d) is out of range",(int)v2->vec.f[0]);
		     yyerror("show where");
		  }
		  if(sm_interrupt) break;
	       }
	    }
	 } else {
	    for(i = 0;i < v1->dimen;i++) {
	       if(set_image_val((int)v2->vec.f[0],(int)v1->vec.f[i],
			     vals->vec.f[0]) < 0) {
		  if(sm_verbose > 1 || first_err) {
		     first_err = 0;
		     msg_1d("IMAGE(%d,",(int)v2->vec.f[0]);
		     msg_1d("%d) is out of range",(int)v1->vec.f[i]);
		     yyerror("show where");
		  }
		  if(sm_interrupt) break;
	       }
	    }
	 }
      } else {
	 if(v1->dimen != vals->dimen) {
	    msg("X and value vectors are different lengths");
	    yyerror("show where");
	    n = (vals->dimen > v1->dimen) ? v1->dimen : vals->dimen;
	 } else {
	    n = v1->dimen;
	 }
	 if(!exchange) {
	    for(i = 0;i < n;i++) {
	       if(set_image_val((int)v1->vec.f[i],(int)v2->vec.f[0],
			     vals->vec.f[i]) < 0) {
		  if(sm_verbose > 1 || first_err) {
		     first_err = 0;
		     msg_1d("IMAGE(%d,",(int)v1->vec.f[i]);
		     msg_1d("%d) is out of range",(int)v2->vec.f[0]);
		     yyerror("show where");
		  }
		  if(sm_interrupt) break;
	       }
	    }
	 } else {
	    for(i = 0;i < n;i++) {
	       if(set_image_val((int)v2->vec.f[0],(int)v1->vec.f[i],
			     vals->vec.f[i]) < 0) {
		  if(sm_verbose > 1 || first_err) {
		     first_err = 0;
		     msg_1d("IMAGE(%d,",(int)v2->vec.f[0]);
		     msg_1d("%d) is out of range",(int)v1->vec.f[i]);
		     yyerror("show where");
		  }
		  if(sm_interrupt) break;
	       }
	    }
	 }
      }
   } else {			/* both vectors */
      if(v1->dimen != v2->dimen) {
	 msg("X and Y vectors are different lengths");
	 yyerror("show where");
      }
      if(vals->dimen == 1) {
	 for(i = 0;i < v2->dimen;i++) {
	    if(set_image_val((int)v1->vec.f[i],(int)v2->vec.f[i],
			     vals->vec.f[0]) < 0) {
	       if(sm_verbose > 1 || first_err) {
		  first_err = 0;
		  msg_1d("IMAGE(%d,",(int)v1->vec.f[i]);
		  msg_1d("%d) is out of range",(int)v2->vec.f[i]);
		  yyerror("show where");
	       }
	       if(sm_interrupt) break;
	    }
	 }
      } else {
	 if(v1->dimen != vals->dimen) {
	    msg("X and value vectors are different lengths");
	    yyerror("show where");
	    n = (vals->dimen > v2->dimen) ? v2->dimen : vals->dimen;
	 } else {
	    n = v2->dimen;
	 }
	 for(i = 0;i < n;i++) {
	    if(set_image_val((int)v1->vec.f[i],(int)v2->vec.f[i],
			     vals->vec.f[i]) < 0) {
	       if(sm_verbose > 1 || first_err) {
		  first_err = 0;
		  msg_1d("IMAGE(%d,",(int)v1->vec.f[i]);
		  msg_1d("%d) is out of range",(int)v2->vec.f[i]);
		  yyerror("show where");
	       }
	       if(sm_interrupt) break;
	    }
	 }
      }
   }
   vec_free(vals);
   vec_free(v1);
   vec_free(v2);
   return;
}

/***********************************************************/
/*
 * Take the sine of a vector
 */
void
vec_sin(v1,ans)
VECTOR *v1,			/* vector to be sined */
       *ans;			/* the answer */
{
   int i;

   if(check_vec(v1,(VECTOR *)NULL,ans) < 0) return;

   for(i = 0;i < v1->dimen;i++) {
      v1->vec.f[i] = sin(v1->vec.f[i]);
   }
   ans->dimen = v1->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v1->vec.f;
   return;
}

/***********************************************************/
/*
 * Take the sqrt of a vector
 */
void
vec_sqrt(v1,ans)
VECTOR *v1,			/* vector to be sqrted */
       *ans;			/* the answer */
{
   int i,
       first_err = 1;			/* First arithmetic error? */

   if(check_vec(v1,(VECTOR *)NULL,ans) < 0) return;

   for(i = 0;i < v1->dimen;i++) {
      if(v1->vec.f[i] >= 0) {
	 v1->vec.f[i] = sqrt(v1->vec.f[i]);
      } else {
	 if(sm_verbose) {
	    if(sm_verbose > 1 || first_err) {
	       first_err = 0;
	       msg("Attempt to take sqrt of non-positive number");
	       yyerror("show where");
	    }
	 }
         if(sm_interrupt) break;
	 v1->vec.f[i] = 0;
      }
   }
   ans->dimen = v1->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v1->vec.f;
   return;
}

/********************************************************/
/*
 * subtract two vectors
 */
void
vec_subtract(v1,v2,ans)
VECTOR *v1,*v2,			/* vectors to be subtracted */
       *ans;			/* the answer */
{
   int i,
       exchange = 0;		/* have vectors been exchanged ? */
   VECTOR *temp;		/* used in exchangeing v1 and v2 */

   if(check_vec(v1,v2,ans) < 0) return;

   if(v1->dimen < v2->dimen) {	/* make v1 >= v2 */
      temp = v1;
      v1 = v2;
      v2 = temp;
      exchange = 1;
   }

   if(v2->dimen == 0) {			/* swap them back, to free longer */
      temp = v1;
      v1 = v2;
      v2 = temp;
   } else if(v2->dimen == 1) {		/* scalar and vector */
      if(exchange) {
	 for(i = 0;i < v1->dimen;i++) {
	    v1->vec.f[i] = v2->vec.f[0] - v1->vec.f[i];
	 }
      } else {
	 for(i = 0;i < v1->dimen;i++) {
	    v1->vec.f[i] -= v2->vec.f[0];
	 }
      }
   } else {			/* both vectors */
      if(v1->dimen != v2->dimen) {
	 msg_2s("%s - %s: vectors' dimensions differ",v1->name,v2->name);
	 yyerror("show where");
      }
      if(exchange) {
	 for(i = 0;i < v2->dimen;i++) {
	    v1->vec.f[i] = v2->vec.f[i] - v1->vec.f[i];
	 }
      } else {
	 for(i = 0;i < v2->dimen;i++) {
	    v1->vec.f[i] -= v2->vec.f[i];
	 }
      }
   }
   vec_free(v2);
   ans->dimen = v1->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v1->vec.f;
   return;
}

/***********************************************************/
/*
 * Take the tangent of a vector
 */
void
vec_tan(v1,ans)
VECTOR *v1,			/* vector to be tangented */
       *ans;			/* the answer */
{
   int i;

   if(check_vec(v1,(VECTOR *)NULL,ans) < 0) return;

   for(i = 0;i < v1->dimen;i++) {
      v1->vec.f[i] = tan(v1->vec.f[i]);
   }
   ans->dimen = v1->dimen;
   ans->type = VEC_FLOAT;
   ans->name = v1->name;
   ans->vec.f = v1->vec.f;
   return;
}

/**********************************************************/
/*
 * Check if v1 (and optionally v2) are arithmetic
 */
int
check_vec(v1,v2,ans)
VECTOR *v1,*v2,*ans;
{
   if(v1->type != VEC_FLOAT || (v2 != NULL && v2->type != VEC_FLOAT)) {
      if(v1->type != VEC_FLOAT) {
	 msg_1s("Vector %s is not arithmetic",v1->name);
	 yyerror("show where");
      }
      if(v2 != NULL && v2->type != VEC_FLOAT) {
	 msg_1s("Vector %s is not arithmetic",v2->name);
	 yyerror("show where");
      }
      vec_free(v1);
      if(v2 != NULL) vec_free(v2);
      if(ans != NULL) {
	 ans->type = VEC_NULL;
	 ans->dimen = 0;
	 ans->name = "(empty)";
      }
      return(-1);
   }
   return(0);
}
